// Copyright (C) 2019  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Package δbtree provides BTree diffing.
package δbtree

//go:generate ../../gen-set δbtree Tree *Tree zset_tree.go

import (
	"fmt"

	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
)

type Tree   = btree.LOBTree
type Bucket = btree.LOBucket

// PathMap is a residency index which maps BTree|Bucket to top tree element.
//
// PathMap is not safe for concurrent access.
type PathMap struct {
	residencyIdx map[zodb.Oid]SetTree // oid -> {} roots
}

// Add adds path to collection.
//
// path[0] signifies a root.
// All path elemens must be Tree except last one which must be Bucket.
func (m *PathMap) Add(path []zodb.IPersistent) {
	l := len(path)
	if l == 0 {
		panic("empty path")
	}

	var root *Tree
	for i, node := range path {
		switch node := node.(type) {
		default:
			panic(fmt.Sprintf("path elements must be Tree|Bucket, not %T", node))

		case *Tree:
			if i == 0 {
				root = node
			}
			// XXX check only not last

		case *Bucket:
			// XXX check only last

		}
	}

	for _, node := range path {
		nodeRoots, ok := m.residencyIdx[node.POid()]
		if !ok {
			nodeRoots = make(SetTree)
			m.residencyIdx[node.POid()] = nodeRoots
		}
		nodeRoots.Add(root)
	}
}

// Invalidates returns which trees are invalidated by changed objects.
//
// returned are: roots of changed trees.
//
// XXX stub.
func (m *PathMap) Invalidates(changev []zodb.Oid) SetTree {
	toinvalidate := make(SetTree)

	for _, oid := range changev {
		roots, ok := m.residencyIdx[oid]
		if !ok {
			continue // we don't know this oid
		}

		toinvalidate.Update(roots)
	}

	return toinvalidate
}
