// Copyright (C) 2018-2019  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main
// misc utilities

import (
	"context"
	"fmt"
	"io"
	"math"
	"syscall"

	log "github.com/golang/glog"

	"github.com/hanwen/go-fuse/fuse"
	"github.com/hanwen/go-fuse/fuse/nodefs"
	"github.com/pkg/errors"

	"lab.nexedi.com/kirr/go123/xcontext"

	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"
)

// ---- FUSE ----

// eInvalError is the error wrapper signifying that underlying error is about "invalid argument".
// err2LogStatus converts such errors into EINVAL return code + logs as warning.
type eInvalError struct {
	err error
}

func (e *eInvalError) Error() string {
	return "invalid argument: " + e.err.Error()
}

// don't propagate eInvalError.Cause -> e.err

func eINVAL(err error) *eInvalError {
	return &eInvalError{err}
}

func eINVALf(format string, argv ...interface{}) *eInvalError {
	return eINVAL(fmt.Errorf(format, argv...))
}

// err2LogStatus converts an error into FUSE status code and logs it appropriately.
//
// the error is logged because otherwise, if e.g. returning EINVAL or EIO
// codes, there is no more detail except the error code itself.
func err2LogStatus(err error) fuse.Status {
	// no error
	if err == nil {
		return fuse.OK
	}

	// direct usage of error code - don't log
	ecode, iscode := err.(syscall.Errno)
	if iscode {
		return fuse.Status(ecode)
	}

	// otherwise log as warnings EINVAL and as errors everything else
	switch errors.Cause(err).(type) {
	case *eInvalError:
		log.Warning(err)
		return fuse.EINVAL

	default:
		log.Error(err)
		return fuse.EIO
	}
}


// asctx represents fuse context as context.Context ready for interrupt handling.
//
// XXX temp. only after proper interrupt handling is not yet merged into go-fuse.
//     https://github.com/hanwen/go-fuse/pull/15
func asctx(fctx *fuse.Context) context.Context {
	// FIXME stub
	return context.Background()
}


// defaultNode should be used instead of nodefs.DefaultNode in wcfs.
//
// nodefs.DefaultNode.Open returns ENOSYS. This is convenient for filesystems
// that have no dynamic files at all. But for filesystems, where there are some
// dynamic files - i.e. nodes which do need to support Open, returning ENOSYS
// from any single node will make the kernel think that the filesystem does not
// support Open at all.
//
// In wcfs we have dynamic files (e.g. /head/watch) and this way we have to
// avoid returning ENOSYS on nodes, that do not need file handles.
//
// defaultNode is like nodefs.defaultNode, but Open returns to kernel fh=0 and
// FOPEN_KEEP_CACHE - similarly how openless case is handled there.
type defaultNode struct{
	nodefs.Node
}

func (n *defaultNode) Open(flags uint32, fctx *fuse.Context) (nodefs.File, fuse.Status) {
	// XXX return something else for directory?
	return &nodefs.WithFlags{
		File:      nil,
		FuseFlags: fuse.FOPEN_KEEP_CACHE,
	}, fuse.OK
}

func newDefaultNode() nodefs.Node {
	return &defaultNode{
		Node: nodefs.NewDefaultNode(),
	}
}


// NewStaticFile creates nodefs.Node for file with static data.
func NewStaticFile(data []byte) *SmallFile {
	return NewSmallFile(func() []byte {
		return data
	})
}

// SmallFile is a nodefs.Node for file with dynamic, but always small, data.
type SmallFile struct {
	nodefs.Node

	// readData gives whole file data
	readData func() []byte
}

func NewSmallFile(readData func() []byte) *SmallFile {
	return &SmallFile{Node: newDefaultNode(), readData: readData}
}

func (f *SmallFile) GetAttr(out *fuse.Attr, _ nodefs.File, _ *fuse.Context) fuse.Status {
	data := f.readData()
	out.Size = uint64(len(data))
	out.Mode = fuse.S_IFREG | 0644
	return fuse.OK
}

func (f *SmallFile) Read(_ nodefs.File, dest []byte, off int64, _ *fuse.Context) (fuse.ReadResult, fuse.Status) {
	data := f.readData()
	l := int64(len(data))
	end := off + l
	if end > l {
		end = l
	}

	return fuse.ReadResultData(data[off:end]), fuse.OK
}


// mkdir adds child to parent as directory.
//
// Note: parent must be already in the filesystem tree - i.e. associated
// with Inode. if not - nodefs will panic in Inode.NewChild on nil dereference.
func mkdir(parent nodefs.Node, name string, child nodefs.Node) {
	parent.Inode().NewChild(name, true, child)
}

// mkfile adds child to parent as file.
//
// Note: parent must be already in the filesystem tree (see mkdir for details).
func mkfile(parent nodefs.Node, name string, child nodefs.Node) {
	parent.Inode().NewChild(name, false, child)
}


// mount is like nodefs.MountRoot but allows to pass in full fuse.MountOptions.
func mount(mntpt string, root nodefs.Node, opts *fuse.MountOptions) (*fuse.Server, *nodefs.FileSystemConnector, error) {
	nodefsOpts := nodefs.NewOptions()
	nodefsOpts.Debug = opts.Debug

	fsconn := nodefs.NewFileSystemConnector(root, nodefsOpts)

	fssrv, err := fuse.NewServer(fsconn.RawFS(), mntpt, opts)
	if err != nil {
		return nil, nil, err
	}

	return fssrv, fsconn, nil
}


// FileSock is bidirectional channel associated with opened file.
//
// FileSock provides streaming write/read operations for filesystem server that
// are correspondingly matched with read/write operations on filesystem user side.
type FileSock struct {
	file *skFile        // filesock's file peer
	rx   *io.PipeReader // socket reads from file here
	tx   *io.PipeWriter // socket writes to file here
}

// skFile is File peer of FileSock.
//
// skFile.Read  is connected with sk.Write.
// skFile.Write is connected with sk.Read.
//
// skFile is always created with nonseekable & directio flags, to support
// streaming semantics.
type skFile struct {
	nodefs.File

	rx *io.PipeReader // file reads from socket here
	tx *io.PipeWriter // file writes to socket here
}

// NewFileSock creates new file socket.
//
// After file socket is created, File return should be given to kernel for the
// socket to be connected to an opened file.
func NewFileSock() *FileSock {
	sk := &FileSock{}
	f  := &skFile{
		File: nodefs.NewDefaultFile(),
	}
	sk.file = f

	rx, tx := io.Pipe()
	sk.rx = rx
	f .tx = tx

	rx, tx = io.Pipe()
	f .rx = rx
	sk.tx = tx

	return sk
}

// File returns nodefs.File handle that is connected to the socket.
//
// The handle should be given to kernel as result of a file open, for that file
// to be connected to the socket.
func (sk *FileSock) File() nodefs.File {
	// nonseekable & directio for opened file to have streaming semantic as
	// if it was a socket.
	return &nodefs.WithFlags{
		File:      sk.file,
		FuseFlags: fuse.FOPEN_NONSEEKABLE | fuse.FOPEN_DIRECT_IO,
	}
}


// Write writes data to filesock.
//
// The data will be read by client reading from filesock's file.
// Write semantic is that of io.Writer.
func (sk *FileSock) Write(data []byte) (n int, err error) {
	// XXX err ctx?
	return sk.tx.Write(data)
}

// Read implements nodefs.File and is paired with filesock.Write().
func (f *skFile) Read(dest []byte, /*ignored*/off int64) (fuse.ReadResult, fuse.Status) {
	n, err := f.rx.Read(dest)
	if n != 0 {
		err = nil
	}
	if err == io.EOF {
		n = 0 // read(2): "zero indicates end of file"
	}
	if err != nil {
		return nil, err2LogStatus(err)
	}
	return fuse.ReadResultData(dest[:n]), fuse.OK
}


// Read reads data from filesock.
//
// The data read will be that the client writes into filesock's file.
// Read semantic is that of io.Reader.
func (sk *FileSock) Read(dest []byte) (n int, err error) {
	// XXX err ctx?
	return sk.rx.Read(dest)
}

// Write implements nodefs.File and is paired with filesock.Read()
func (f *skFile) Write(data []byte, /*ignored*/off int64) (uint32, fuse.Status) {
	// cap data to 2GB (not 4GB not to overflow int on 32-bit platforms)
	l := len(data)
	if l > math.MaxInt32 {
		l = math.MaxInt32
		data = data[:l]
	}

	n, err := f.tx.Write(data)
	if n != 0 {
		err = nil
	}
	if err == io.ErrClosedPipe {
		err = syscall.ECONNRESET // XXX better EPIPE?
	}
	if err != nil {
		return 0, err2LogStatus(err)
	}
	return uint32(n), fuse.OK
}


// CloseRead closes reading side of filesock.
func (sk *FileSock) CloseRead() error {
	return sk.rx.Close()
}

// CloseWrite closes writing side of filesock.
func (sk *FileSock) CloseWrite() error {
	return sk.tx.Close()
}

// Close closes filesock.
//
// it is semantically equivalent to CloseRead + CloseWrite.
func (sk *FileSock) Close() error {
	err := sk.CloseRead()
	err2 := sk.CloseWrite()
	if err == nil {
		err = err2
	}
	return err
}

// Flush implements nodefs.File to handle close(file) call.
func (f *skFile) Flush() fuse.Status {
	err := f.rx.Close()
	err2 := f.tx.Close()
	if err == nil {
		err = err2
	}
	return err2LogStatus(err)
}


// ---- ZODB ---

// typeOf returns string for object's type.
//
//	- for ZODB objects, it uses zodb.ClassOf, which in particular supports
//	  printing zodb.Broken with details properly.
//
//	- for other objects, it uses %T.
func typeOf(obj interface{}) string {
	switch obj := obj.(type) {
	case zodb.IPersistent:
		return zodb.ClassOf(obj)

	default:
		return fmt.Sprintf("%T", obj)
	}
}

// ZConn is zodb.Connection + associated read-only transaction under which
// objects of the connection are accessed.
type ZConn struct {
	*zodb.Connection

	// read-only transaction under which we access zodb.Connection data.
	txnCtx context.Context // XXX -> better directly store txn

/* XXX redecided to purge files and @<rev>/ on atime timeout

	// for historic @<rev> access the connection can be shared between several bigfiles.
	// since we want to free such connections when no longer needed we
	// return zodb.Connection back to zodb.DB when refcnt drops to 0.
	refcnt int32
*/
}

// zopen opens new connection to ZODB database + associated read-only transaction.
func zopen(ctx context.Context, zdb *zodb.DB, zopt *zodb.ConnOptions) (_ *ZConn, err error) {
	// create new read-only transaction
	txn, txnCtx := transaction.New(context.Background())
	defer func() {
		if err != nil {
			txn.Abort()
		}
	}()

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(ctx, txnCtx)
	defer cancel()

	zconn, err := zdb.Open(ctx, zopt)
	if err != nil {
		return nil, err
	}

	return &ZConn{
		Connection: zconn,
		txnCtx:     txnCtx,
//		refcnt:     1,
	}, nil
}

/*
// Release decrements reference count and releases connection back to zodb.DB
// if it is no longer used.
func (zc *ZConn) Release() {
	rc := atomic.AddInt32(&zc.refcnt, -1)
	if rc < 0 {
		panic("ZConn.Release: refcnt < 0")
	}
	if rc > 0 {
		return
	}

	txn := transaction.Current(zc.txnCtx)
	txn.Abort()

	zc.Connection = nil
	zc.txnCtx = nil
}

// Incref increments connection's reference counter by 1.
func (zc *ZConn) Incref() {
	rc := atomic.AddInt32(&zc.refcnt, +1)
	if rc <= 1 {
		panic("Zconn.Incref: refcnt was < 1")
	}
}


// zopenAt opens historic connection @<rev>.
//
// if the connection for this @<rev> was already opened - it is shared.
func (r *Root) zopenAt(ctx context.Context, rev zodb.Tid) (_ *ZConn, err error) {
	// check if zconn for @<rev> is already there
	r.revMu.Lock()
	zconn := r.revTab[rev]
	if zconn != nil {
		if atomic.LoadInt32(&zconn.refcnt) > 0 {
			zconn.Incref()
		} else {
			zconn = nil // in-progress destruction
		}
	}
	r.revMu.Unlock()

	if zconn != nil {
		return zconn, nil
	}

	// not there - without revMu lock proceed to open it
	zconn, err = zopen(ctx, r.zdb, &zodb.ConnOptions{At: rev})
	if err != nil {
		return nil, err
	}

	// relock revTab and either register zconn, or returun another zconn2,
	// that might have been opened while we were not holding revMu.
	r.revMu.Lock()
	defer r.revMu.Unlock()

	zconn2 := r.revTab[rev]
	if zconn2 != nil {
		if atomic.LoadInt32(&zconn2.refcnt) > 0 {
			zconn.Release()	// FIXME aborts txn -> just drop zconn
			zconn2.Incref()
			return zconn2, nil
		}
		// else it was another in-progress destruction
	}

	r.revTab[rev] = zconn

	// schedule del revTab[rev] on zconn destroy
	txn := transaction.Current(zconn.txnCtx)
	txn.RegisterSync(&zrevTabUnregister{r, zconn})

	return zconn, nil
}

// zrevTabUnregister unregisters zconn from root.revTab on zconn's transaction abort.
type zrevTabUnregister struct {
	root  *Root
	zconn *ZConn
}

func (u *zrevTabUnregister) BeforeCompletion(txn transaction.Transaction) {}
func (u *zrevTabUnregister) AfterCompletion(txn transaction.Transaction) {
	rev := u.zconn.At()

	u.root.revMu.Lock()
	defer u.root.revMu.Unlock()

	// delete only if zconn is still registered - as another zconn2 might have
	// been already registered instead while zconn was in revTab with refcnt=0.
	if u.root.revTab[rev] == u.zconn {
		delete(u.root.revTab, rev)
	}
}
*/
