// Copyright (C) 2018-2019  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Program wcfs provides filesystem server with file data backed by wendelin.core arrays.
//
// Intro
//
// Each wendelin.core array (ZBigArray) is actually a linear file (ZBigFile)
// and array metadata like dtype, shape and strides associated with it. This
// program exposes as files only ZBigFile data and leaves rest of
// array-specific handling to clients. Every ZBigFile is exposed as one separate
// file that represents whole ZBigFile's data.
//
// For a client, the primary way to access a bigfile should be to mmap
// head/bigfile/<bigfileX> which represents always latest bigfile data.
// Clients that want to get isolation guarantee should subscribe for
// invalidations and re-mmap invalidated regions to file with pinned bigfile revision for
// the duration of their transaction. See "Invalidation protocol" for details.
//
// In the usual situation when bigfiles are big, and there are O(1)/δt updates,
// there should be no need for any cache besides shared kernel cache of latest
// bigfile data.
//
//
// Filesystem organization
//
// Top-level structure of provided filesystem is as follows:
//
//	head/			; latest database view
//		...
//	@<rev1>/		; database view as of revision <revX>
//		...
//	@<rev2>/
//		...
//	...
//
// where head/ represents latest data as stored in upstream ZODB, and
// @<revX>/ represents data as of database revision <revX>.
//
// head/ has the following structure:
//
//	head/
//		at			; data inside head/ is as of this ZODB transaction
//		watch			; channel for bigfile invalidations
//		bigfile/		; bigfiles' data
//			<oid(bigfile1)>
//			<oid(bigfile2)>
//			...
//
// where /bigfile/<bigfileX> represents latest bigfile data as stored in
// upstream ZODB. As there can be some lag receiving updates from the database,
// /at describes precisely ZODB state for which bigfile data is currently
// exposed. Whenever bigfile data is changed in upstream ZODB, information
// about the changes is first propagated to /watch, and only after that
// /bigfile/<bigfileX> is updated. See "Invalidation protocol" for details.
//
// @<revX>/ has the following structure:
//
//	@<revX>/
//		at
//		bigfile/		; bigfiles' data as of revision <revX>
//			<oid(bigfile1)>
//			<oid(bigfile2)>
//			...
//
// where /bigfile/<bigfileX> represent bigfile data as of revision <revX>.
//
// Unless accessed {head,@<revX>}/bigfile/<bigfileX> are not automatically visible in
// wcfs filesystem. Similarly @<revX>/ should be explicitly created by client via mkdir.	XXX -> just @<revX> access.
//
//
// Invalidation protocol
//
// In order to support isolation, wcfs implements invalidation protocol that
// must be cooperatively followed by both wcfs and client.
//
// First, client mmaps latest bigfile, but does not access it
//
//	mmap(head/bigfile/<bigfileX>)
//
// Then client opens head/watch and tells wcfs through it for which ZODB state
// it wants to get bigfile's view.
//
//	C: 1 watch <bigfileX> @<at>
//
// The server then, after potentially sending initial pin messages (see below),
// reports either success or failure:
//
//	S: 1 ok
//	S: 1 error ...		; if <at> is too far away back from head/at
//
// The server sends "ok" reply only after head/at is ≥ requested <at>, and
// only after all initial pin messages are fully acknowledged by the client.
// The client can start to use mmapped data after it gets "ok".
// The server sends "error" reply if requested <at> is too far away back from
// head/at.
//
// Upon watch request, either initially, or after sending "ok", the server will be notifying the
// client about file blocks that client needs to pin in order to observe file's
// data as of <at> revision:
//
// The filesystem server itself receives information about changed data from
// ZODB server through regular ZODB invalidation channel (as it is ZODB client
// itself). Then, separately for each changed file block, before actually
// updating head/bigfile/<bigfileX> content, it notifies through head/watch to
// clients, that had requested it (separately to each client), about the
// changes:
//
//	S: 2 pin <bigfileX> #<blk> @<rev_max>
//
// and waits until all clients confirm that changed file block can be updated
// in global OS cache.
//
// The client in turn should now re-mmap requested to be pinned block to bigfile@<rev_max>
//
//	# mmapped at address corresponding to #blk
//	mmap(@<rev_max>/bigfile/<bigfileX>, #blk, MAP_FIXED)
//
// and must send ack back to the server when it is done:
//
//	C: 2 ack
//
// The server sends pin notifications only for file blocks, that are known to
// be potentially changed after client's <at>, and <rev_max> describes the
// upper bound for the block revision:
//
//	<at>	<  <rev_max>		FIXME ->	<rev_max> ≤ <at>	(?)
//
// The server maintains short history tail of file changes to be able to
// support openings with <at> being slightly in the past compared to current
// head/at. The server might reject a watch request if <at> is too far away in
// the past from head/at. The client is advised to restart its transaction with
// more uptodate database view if it gets watch setup error.
//
// A later request from the client for the same <bigfileX> but with different
// <at>, overrides previous watch request for that file. A client can use "-"
// instead of "@<at>" to stop watching a file.
//
// A single client can send several watch requests through single head/watch
// open, as well as it can use several head/watch opens simultaneously.
// The server sends pin notifications for all files requested to be watched via
// every head/watch open.
//
// Note: a client could use a single watch to manage its several views for the same
// file but with different <at>. This could be achieved via watching with
// @<at_min>, and then deciding internally which views needs to be adjusted and
// which views need not. Wcfs does not oblige clients to do so though, and a
// client is free to use as many head/watch openenings as it needs to.
//
// When clients are done with @<revX>/bigfile/<bigfileX> (i.e. client's
// transaction ends and array is unmapped), the server sees number of opened
// files to @<revX>/bigfile/<bigfileX> drops to zero, and automatically
// destroys @<revX>/bigfile/<bigfileX> after reasonable timeout.
//
//
// Protection against slow or faulty clients
//
// If a client, on purpose or due to a bug or being stopped, is slow to respond
// with ack to file invalidation notification, it creates a problem because the
// server will become blocked waiting for pin acknowledgments, and thus all
// other clients, that try to work with the same file, will get stuck.
//
// The problem could be avoided, if wcfs would reside inside OS kernel and this
// way could be able to manipulate clients address space directly (then
// invalidation protocol won't be needed). It is also possible to imagine
// mechanism, where wcfs would synchronously change clients' address space via
// injecting trusted code and running it on client side via ptrace to adjust
// file mappings.
//
// However ptrace does not work when client thread is blocked under pagefault,
// and that is exactly what wcfs would need to do to process invalidations
// lazily, because eager invalidation processing results in prohibitively slow
// file opens. See internal wcfs overview for details about why ptrace
// cannot be used and why lazy invalidation processing is required.
//
// Lacking OS primitives to change address space of another process and not
// being able to work it around with ptrace in userspace, wcfs takes approach
// to kill a slow client on 30 seconds timeout by default.
//
//
// Writes
//
// As each bigfile is represented by 1 synthetic file, there can be several
// write schemes:
//
// 1. mmap(MAP_PRIVATE) + writeout by client
//
// In this scheme bigfile data is mmapped in MAP_PRIVATE mode, so that local
// user changes are not automatically propagated back to the file. When there
// is a need to commit, client investigates via some OS mechanism, e.g.
// /proc/self/pagemap or something similar, which pages of this mapping it
// modified. Knowing this it knows which data it dirtied and so can write this
// data back to ZODB itself, without filesystem server providing write support.
//
// 2. mmap(MAP_SHARED, PROT_READ) + write-tracking & writeout by client
//
// In this scheme bigfile data is mmaped in MAP_SHARED mode with read-only pages
// protection. Then whenever write fault occurs, client allocates RAM from
// shmfs, copies faulted page to it, and then mmaps RAM page with RW protection
// in place of original bigfile page. Writeout implementation should be similar
// to "1", only here client already knows the pages it dirtied, and this way
// there is no need to consult /proc/self/pagemap.
//
// The advantage of this scheme over mmap(MAP_PRIVATE) is that in case
// there are several in-process mappings of the same bigfile with overlapping
// in-file ranges, changes in one mapping will be visible in another mapping.
// Contrary: whenever a MAP_PRIVATE mapping is modified, the kernel COWs
// faulted page into a page completely private to this mapping, so that other
// MAP_PRIVATE mappings of this file, including ones created from the same
// process, do not see changes made to the first mapping.
//
// Since wendelin.core needs to provide coherency in between different slices
// of the same array, this is the mode wendelin.core actually uses.
//
// 3. write to wcfs
//
// TODO we later could implement "write-directly" mode where clients would write
// data directly into the file.
package main

// Wcfs organization
//
// Wcfs is a ZODB client that translates ZODB objects into OS files as would
// non-wcfs wendelin.core do for a ZBigFile. Contrary to non-wcfs wendelin.core,
// it keeps bigfile data in shared OS cache efficiently. It is organized as follows:
//
// 1) 1 ZODB connection for "latest data" for whole filesystem (zhead).
// 2) head/bigfile/* of all bigfiles represent state as of zhead.At .
// 3) for head/bigfile/* the following invariant is maintained:
//
//	#blk ∈ file cache    =>    ZBlk(#blk) + all BTree/Bucket that lead to it  ∈ zhead cache
//	                           (ZBlk* in ghost state(%))
//
//    The invariant helps on invalidation: if we see a changed oid, and
//    zhead.cache.lookup(oid) = ø -> we know we don't have to invalidate OS
//    cache for any part of any file (even if oid relates to a file block - that
//    block is not cached and will trigger ZODB load on file read).
//
//    Currently we maintain this invariant by simply never evicting LOBTree/LOBucket
//    objects from ZODB Connection cache (LOBucket keeps references to ZBlk* and
//    so ZBlk* also stay in cache in ghost form). In the future we may want to
//    try to synchronize to kernel freeing its pagecache pages.
//
// 4) when we receive an invalidation message from ZODB - we process it and
//    propagate invalidations to OS file cache of head/bigfile/*:
//
//	invalidation message: (tid↑, []oid)
//
//    4.1) zhead.cache.lookup(oid)
//    4.2) ø: nothing to do - see invariant ^^^.
//    4.3) obj found:
//
//	- ZBlk*		-> file/#blk
//	- BTree/Bucket	-> δ(BTree)  -> file/[]#blk
//
//	in the end after processing all []oid from invalidation message we have
//
//	  [] of file/[]#blk
//
//	that describes which file(s) parts needs to be invalidated.
//
//    4.4) for all file/blk to invalidate we do:
//
//	- try to retrieve head/bigfile/file[blk] from OS file cache(*);
//	- if retrieved successfully -> store retrieved data back into OS file
//	  cache for @<rev>/bigfile/file[blk], where
//
//	    rev = max(δFtail.by(#blk)) || min(rev ∈ δFtail) || zhead.at	; see below about δFtail
//
//	- invalidate head/bigfile/file[blk] in OS file cache.
//
//	This preserves previous data in OS file cache in case it will be needed
//	by not-yet-uptodate clients, and makes sure file read of head/bigfile/file[blk]
//	won't be served from OS file cache and instead will trigger a FUSE read
//	request to wcfs.
//
//    4.5) no invalidation messages are sent to wcfs clients at this point(+).
//
//    4.6) processing ZODB invalidations and serving file reads (see 7) are
//      organized to be mutually exclusive.
//
//	(TODO head.zconnMu -> special mutex with Lock(ctx) so that Lock could be canceled)
//
// 5) after OS file cache was invalidated, we resync zhead to new database
//    view corresponding to tid.
//
// 6) for every file δFtail invalidation info about head/data is maintained:
//
//	- tailv: [](rev↑, []#blk)
//	- by:    {} #blk -> []rev↑ in tail
//
//    δFtail.tail describes invalidations to file we learned from ZODB invalidation.
//    δFtail.by   allows to quickly lookup information by #blk.
//
//    min(rev) in δFtail is min(@at) at which head/bigfile/file is currently mmapped (see below).
//
//    to support initial openings with @at being slightly in the past, we also
//    make sure that min(rev) is enough to cover last 10 minutes of history
//    from head/at.
//
// 7) when we receive a FUSE read(#blk) request to a head/bigfile/file we process it as follows:
//
//   7.1) load blkdata for head/bigfile/file[blk] @zhead.at .
//
//	while loading this also gives upper bound estimate of when the block
//	was last changed:
//
//	  rev(blk) ≤ max(_.serial for _ in (ZBlk(#blk), all BTree/Bucket that lead to ZBlk))
//
//	it is not exact because BTree/Bucket can change (e.g. rebalance)
//	but still point to the same k->ZBlk.
//
//	we also use file.δFtail to find either exact blk revision:
//
//	  rev(blk) = max(file.δFtail.by(#blk) -> []rev↑)
//
//	or another upper bound if #blk ∉ δFtail:
//
//	  rev(blk) ≤ min(rev ∈ δFtail)		; #blk ∉ δFtail
//
//
//	below rev'(blk) is min(of the estimates found):
//
//	  rev(blk) ≤ rev'(blk)		rev'(blk) = min(^^^)
//
//
//   7.2) for all client@at mmappings of head/bigfile/file:
//
//	- rev'(blk) ≤ at: -> do nothing		XXX if rev is not last and there is rev_next ≤ at ? -> also consult δFtail ?
//	- rev'(blk) > at:
//	  - if blk ∈ mmapping.pinned -> do nothing
//	  - rev = max(δFtail.by(#blk) : _ ≤ at)	|| min(rev ∈ δFtail : rev ≤ at)	|| at
//	  - client.remmap(file, #blk, @rev/bigfile/file)
//	  - mmapping.pinned += blk
//
//	remmapping is done via "invalidation protocol" exchange with client.
//	( one could imagine adjusting mappings synchronously via running
//	  wcfs-trusted code via ptrace that wcfs injects into clients, but ptrace
//	  won't work when client thread is blocked under pagefault or syscall(^) )
//
//	in order to support remmapping for each head/bigfile/file
//
//	  [] of mmapping{client@at↑, pinned}
//
//	is maintained.
//
//   7.3) blkdata is returned to kernel.
//
//   Thus a client that wants latest data on pagefault will get latest data,
//   and a client that wants @rev data will get @rev data, even if it was this
//   "old" client that triggered the pagefault(~).
//
// XXX 8) serving read from @<rev>/data + zconn(s) for historical state
// XXX 9) gc @rev/ and @rev/bigfile/<bigfileX> automatically on atime timeout
//
//
// (*) see notes.txt -> "Notes on OS pagecache control"
// (+) see notes.txt -> "Invalidations to wcfs clients are delayed until block access"
// (~) see notes.txt -> "Changing mmapping while under pagefault is possible"
// (^) see notes.txt -> "Client cannot be ptraced while under pagefault"
// (%) no need to keep track of ZData - ZBlk1 is always marked as changed on blk data change.
//
// XXX For every ZODB connection a dedicated read-only transaction is maintained.

import (
	"context"
	"flag"
	"fmt"
	stdlog "log"
	"os"
	"runtime"
	"strings"
	"sync"
	"syscall"

	log "github.com/golang/glog"
	"golang.org/x/sync/errgroup"

	"lab.nexedi.com/kirr/go123/xcontext"
	"lab.nexedi.com/kirr/go123/xerr"

	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
	_ "lab.nexedi.com/kirr/neo/go/zodb/wks"

	"github.com/hanwen/go-fuse/fuse"
	"github.com/hanwen/go-fuse/fuse/nodefs"
	"github.com/pkg/errors"

	"./internal/δbtree"
)

// Root represents root of wcfs filesystem.
type Root struct {
	nodefs.Node

	// ZODB storage we work with
	zstor zodb.IStorage

	// ZODB DB handle for zstor.
	// keeps cache of connections for @<rev>/ accesse.
	// only one connection is used for each @<rev>.
	zdb *zodb.DB

	// directory + ZODB connection for head/
	// (zhead is Resync'ed and is kept outside zdb pool)
	head *Head

	// directories + ZODB connections for @<rev>/
	revMu  sync.Mutex
	revTab map[zodb.Tid]*Head
}

// /(head|<rev>)/			- served by Head.
type Head struct {
	nodefs.Node
	rev   zodb.Tid    // 0 for head/, !0 for @<rev>/
	bfdir *BigFileDir // bigfile/
	// at    - served by .readAt
	// watch - implicitly linked to by fs

	// ZODB connection for everything under this head
	zconnMu sync.RWMutex // protects access to zconn & live _objects_ associated with it
	zconn   *ZConn       // for head/ zwatcher resyncs head.zconn; others only read zconn objects.

	// XXX move zconn's current transaction to Head here?

}

// /head/watch				- served by Watch.
type Watch struct {
	nodefs.Node

	// TODO
}

// /(head|<rev>)/bigfile/		- served by BigFileDir.
type BigFileDir struct {
	nodefs.Node
	head *Head // parent head/ or @<rev>/

	// {} oid -> <bigfileX>
	mu      sync.Mutex
	fileTab map[zodb.Oid]*BigFile

	// residency index for btree|bucket -> which file
	btreeMap *δbtree.PathMap
}

// /(head|<rev>)/bigfile/<bigfileX>	- served by BigFile.
type BigFile struct {
	nodefs.Node

	// this BigFile is under .head/bigfile/; it views ZODB via .head.zconn
	// parent's BigFileDir.head is the same.
	head	*Head

	// ZBigFile top-level object. Kept activated during lifetime of current transaction.
	zbf	*ZBigFile

	// zbf.Size(). It is constant during liftime of current transaction.
	zbfSize int64

	// tail change history of this file.
	δFtail *ΔTailI64 // [](rev↑, []#blk)

	// inflight loadings of ZBigFile from ZODB.
	// successfull load results are kept here until blkdata is put into OS pagecache.
	loadMu  sync.Mutex
	loading map[int64]*blkLoadState // #blk -> {... blkdata}

	// XXX mappings where client(s) requested isolation guarantee
	//mappings ...
}

// blkLoadState represents a ZBlk load state/result.
//
// when !ready the loading is in progress.
// when ready the loading has been completed.
type blkLoadState struct {
	ready chan struct{}

	blkdata  []byte
	err      error
}

// ----------------------------------------

// zodbCacheControl implements zodb.LiveCacheControl to tune ZODB to never evict
// LOBTree/LOBucket from live cache. We want to keep LOBTree/LOBucket always alive
// becuse it is essentially the index where to find ZBigFile data.
//
// For the data itself - we put it to kernel pagecache and always deactivate
// from ZODB right after that.
//
// See "3) for */head/data the following invariant is maintained..."
type zodbCacheControl struct {}

func (cc *zodbCacheControl) WantEvict(obj zodb.IPersistent) bool {
	switch obj.(type) {
	default:
		return true

	case *btree.LOBTree:
	case *btree.LOBucket:

	// ZBlk* are kept referenced by a LOBucket, so they don't go away from Connection.cache.objtab

	// FIXME we need to keep ZBigFile in cache: even if we keep a pointer
	// to ZBigFile, but it is allowed to drop its state, it will release
	// pointer to LOBTree object and, consequently, that LOBTree object,
	// even if it was marked not to be released from cache will be GC'ed by
	// go runtime, and the cache will loose its weak reference to it.
	// XXX however we cannot protect ZBigFile from releaseing state - as
	// any object can be explicitly invalidated.
	// FIXME -> teach zodb.LiveCache to keep object by itself?
	//
	// we also keep ZBigFile alive because we want to make sure .blksize
	// and (p. ref) .blktab do not change.
	//
	// XXX on every resynce we deactivate/activate all bigfiles and restat them
	//     -> for efficiency better keep ZBigFile in live cache.
	//case *ZBigFile:
	}

	return false
}

func traceWatch(format string, argv ...interface{}) {
	if !log.V(1) {	// XXX -> 2?
		return
	}

	log.Infof("zwatcher: " + format, argv...)	// XXX InfoDepthf
}

// zwatcher watches for ZODB changes.
//
// see "4) when we receive an invalidation message from ZODB ..."
func (root *Root) zwatcher(ctx context.Context) (err error) {
	defer xerr.Contextf(&err, "zwatch")	// XXX more in context?
	// XXX unmount on error? -> always EIO?

	traceWatch(">>>")

	zwatchq := make(chan zodb.CommitEvent)
	at0 := root.zstor.AddWatch(zwatchq)	// XXX -> to main thread to avoid race
	defer root.zstor.DelWatch(zwatchq)
	_ = at0 // XXX

	var zevent zodb.CommitEvent
	var ok bool

	for {
		traceWatch("select ...")
		select {
		case <-ctx.Done():
			traceWatch("cancel")
			return ctx.Err()

		// TODO handle errors from ZODB watch stream
		case zevent, ok = <-zwatchq:
			if !ok {
				traceWatch("zwatchq closed")
				return nil // closed	XXX ok?
			}

			traceWatch("zevent: %s", zevent)
		}

		root.zδhandle1(zevent)
	}
}

// zδhandle1 handles 1 change event from ZODB notification.
func (root *Root) zδhandle1(zevent zodb.CommitEvent) {
	// while we are invalidating OS cache, make sure that nothing, that
	// even reads /head/bigfile/*, is running (see 4.6).
	root.head.zconnMu.Lock()
	defer root.head.zconnMu.Unlock()

	zhead := root.head.zconn
	bfdir := root.head.bfdir

	toinvalidate := map[*BigFile]SetI64{} // {} file -> set(#blk)

	fmt.Printf("\n\n\n")

	// zevent = (tid↑, []oid)
	for _, oid := range zevent.Changev {
		// XXX zhead.Cache() lock/unlock + comment it is not really needed
		obj := zhead.Cache().Get(oid)
		if obj == nil {
			fmt.Printf("%s: not in cache\n", oid)
			continue // nothing to do - see invariant
		}

		fmt.Printf("%s:     in cache (%s)\n", oid, typeOf(obj))

		switch obj := obj.(type) {
		default:
			continue // object not related to any bigfile

		case *btree.LOBTree:
			// XXX -> δBTree

		case *btree.LOBucket:
			// XXX -> δBTree

		case zBlk:	// ZBlk*
			// blkBoundTo locking: no other bindZFile are running,
			// since we write-locked head.zconnMu and bindZFile is
			// run when loading objects - thus when head.zconnMu is
			// read-locked.
			//
			// bfdir locking: similarly not needed, since we are
			// exclusively holding head lock.
			for zfile, objBlk := range obj.blkBoundTo() {
				file, ok := bfdir.fileTab[zfile.POid()]
				if !ok {
					// even though zfile is in ZODB cache, the
					// filesystem already forgot about this file.
					continue
				}

				blkmap, ok := toinvalidate[file]
				if !ok {
					blkmap = SetI64{}
					toinvalidate[file] = blkmap
				}
				blkmap.Update(objBlk)
			}

		case *ZBigFile:
			// XXX check that .blksize and .blktab (it is only
			// persistent reference) do not change.

			// XXX shutdown fs with ^^^ message.
		}

		// make sure obj won't be garbage-collected until we finish handling it.
		runtime.KeepAlive(obj)
	}

	fmt.Printf("\n\nzδhandle: toinvalidate (#%d):\n", len(toinvalidate))
	for file := range toinvalidate {
		fmt.Printf("\t- %s\n", file.zbf.POid())
	}

	wg, ctx := errgroup.WithContext(context.TODO())
	for file, blkmap := range toinvalidate {
		file := file
		for blk := range blkmap {
			blk := blk
			wg.Go(func() error {
				return file.invalidateBlk(ctx, blk)
			})
		}

		// invalidate kernel cache for attributes
		// XXX we need to do it only if we see topoligy (i.e. btree) change
		wg.Go(func() error {
			return file.invalidateAttr()
		})
	}
	err := wg.Wait()
	if err != nil {
		panic(err)	// XXX
	}

	// resync .zhead to zevent.tid
	// XXX -> Head.Resync() ?

	// 1. deactivate all ZBigFile (we keep them activated during whole txn)
	// XXX dir.mu locking (not needed bcause zconnMu locked)
	for _, file := range bfdir.fileTab {
		file.zbf.PDeactivate()
		// XXX we need to reread size only for files in toinvalidate
		file.zbfSize = -1 // just in case
	}

	// 2. abort old and resync to new txn/at
	transaction.Current(zhead.txnCtx).Abort()
	txn, ctx := transaction.New(context.Background()) // XXX bg ok?
	zhead.Resync(txn, zevent.Tid)
	zhead.txnCtx = ctx

	// 3. reactivate/restat all ZBigFile
	// XXX -> parallel?
	// XXX locking
	for _, file := range bfdir.fileTab {
		err := file.zbf.PActivate(ctx)
		if err != nil {
			panic(err)	// XXX
		}

		zbfSize, err := file.zbf.Size(ctx)
		if err != nil {
			panic(err)	// XXX
		}

		file.zbfSize = zbfSize
	}

	// notify .wcfs/zhead
	for sk := range gdebug.zheadSockTab {
		_, err := fmt.Fprintf(sk, "%s\n", zevent.Tid)
		if err != nil {
			log.Error(err)	// XXX errctx, -> warning?
			sk.Close()
			delete(gdebug.zheadSockTab, sk)
		}
	}
}

// invalidateBlk invalidates 1 file block in kernel cache.
//
// see "4.4) for all file/blk to in invalidate we do"
//
// called with f.head.zconnMu wlocked.
func (f *BigFile) invalidateBlk(ctx context.Context, blk int64) error {
	fsconn := gfsconn
	blksize := f.zbf.blksize
	off := blk*blksize

	// try to retrieve cache of current head/data[blk]
	//
	// if less than blksize was cached - probably the kernel had to evict
	// some data from its cache already. In such case we don't try to
	// preserve the rest and drop what was read, to avoid keeping the
	// system overloaded.
	//
	// XXX st != OK -> warn?
	blkdata := make([]byte, blksize)
	n, st := fsconn.FileRetrieveCache(f.Inode(), off, blkdata)
	if int64(n) == blksize {
		// XXX -> go
		// store retrieved data back to OS cache for file @<rev>/file[blk]
		blkrev, _ := f.δFtail.LastRevOf(blk, f.head.zconn.At())
		frev, err := groot.mkrevfile(blkrev, f.zbf.POid())
		if err != nil {
			// XXX
		}
		st = fsconn.FileNotifyStoreCache(frev.Inode(), off, blkdata)
		if st != fuse.OK {
			// XXX log	- dup wrt readBlk -> common func.
		}
	}

	// invalidate file/head/data[blk] in OS file cache.
	st = fsconn.FileNotify(f.Inode(), off, blksize)
	// XXX st != ok (fatal here)

	panic("TODO")
}

// invalidateAttr invalidates file attributes in kernel cache.
func (f *BigFile) invalidateAttr() error {
	log.Error("AAA")

	fsconn := gfsconn
	st := fsconn.FileNotify(f.Inode(), -1, -1) // metadata only
	// st != ok
	_ = st
	panic("TODO")
}


// mkrevfile makes sure inode ID of /@<rev>/bigfile/<fid> is known to kernel.
//
// We need node ID to be know to the kernel, when we need to store data into
// file's kernel cache - if the kernel don't have the node ID for the file in
// question, FileNotifyStoreCche will just fail.
//
// For kernel to know the inode mkrevfile issues regular filesystem lookup
// request which goes to kernel and should go back to wcfs. It is thus not safe
// to use mkrevfile from under FUSE request handler as doing so might deadlock.
func (root *Root) mkrevfile(rev zodb.Tid, fid zodb.Oid) (_ *BigFile, err error) {
	fsconn := gfsconn

	frevpath := fmt.Sprintf("@%s/bigfile/%s", rev, fid) // relative to fs root for now
	defer xerr.Contextf(&err, "/: mkrevfile %s", frevpath)

	// first check without going through kernel, whether the inode maybe know already
	xfrev := fsconn.LookupNode(root.Inode(), frevpath)
	if xfrev != nil {
		// FIXME checking for "node{0}" is fragile, but currently no other way
		if xfrev.String() != "node{0}" {
			return xfrev.Node().(*BigFile), nil
		}
	}

	// we have to ping the kernel
	frevospath := gmntpt + "/" + frevpath // now starting from OS /
	f, err := os.Open(frevospath)
	if err != nil {
		return nil, err
	}
	defer f.Close()

	xfrev = fsconn.LookupNode(root.Inode(), frevpath)
	// must be !nil as open succeeded	XXX better recheck
	return xfrev.Node().(*BigFile), nil
}

// ----------------------------------------

// /(head|<rev>)/at -> readAt serves read.
func (h *Head) readAt() []byte {
	h.zconnMu.RLock()
	defer h.zconnMu.RUnlock()

	return []byte(h.zconn.At().String())
}

// /(head|<rev>)/bigfile/ -> Lookup receives client request to create /(head|<rev>)/bigfile/<bigfileX>.
func (bfdir *BigFileDir) Lookup(out *fuse.Attr, name string, fctx *fuse.Context) (*nodefs.Inode, fuse.Status) {
	f, err := bfdir.lookup(out, name, fctx)
	var inode *nodefs.Inode
	if f != nil {
		inode = f.Inode()
	}
	return inode, err2LogStatus(err)

}

func (bfdir *BigFileDir) lookup(out *fuse.Attr, name string, fctx *fuse.Context) (f *BigFile, err error) {
	defer xerr.Contextf(&err, "/XXXbigfile: lookup %q", name)	// XXX name -> path

	oid, err := zodb.ParseOid(name)
	if err != nil {
		return nil, eINVALf("not oid")
	}

	bfdir.head.zconnMu.RLock()
	defer bfdir.head.zconnMu.RUnlock()

	defer func() {
		if f != nil {
			f.getattr(out)
		}
	}()

	// check to see if dir(oid) is already there
	bfdir.mu.Lock()
	f, already := bfdir.fileTab[oid]
	bfdir.mu.Unlock()

	if already {
		return f, nil
	}

	// not there - without bfdir lock proceed to open BigFile from ZODB
	f, err = bfdir.head.bigopen(asctx(fctx), oid)
	if err != nil {
		return nil, err
	}

	// relock bfdir and either register f or, if the file was maybe
	// simultanously created while we were not holding bfdir.mu, return that.
	bfdir.mu.Lock()
	f2, already := bfdir.fileTab[oid]
	if already {
		bfdir.mu.Unlock()
		f.Close()
		return f2, nil
	}

	bfdir.fileTab[oid] = f
	bfdir.mu.Unlock()

	// mkfile takes filesystem treeLock - do it outside bfdir.mu
	mkfile(bfdir, name, f)

	return f, nil
}

// XXX do we need to support unlink? -> no, @<revX>/ are automatically garbage-collected.

// / -> Mkdir receives client request to create @<rev>/.
//
// it is not an error if @<rev>/ already exists - mkdir succeeds and EEXIST is not returned.
// in other words mkdir behaves here similarly to `mkdir -p`.
//
// XXX -> remove mkdir and just create @revX/ on lookup.
func (root *Root) Mkdir(name string, mode uint32, fctx *fuse.Context) (*nodefs.Inode, fuse.Status) {
	inode, err := root.mkdir(name, fctx) // XXX ok to ignore mode?
	return inode, err2LogStatus(err)
}

func (root *Root) mkdir(name string, fctx *fuse.Context) (_ *nodefs.Inode, err error) {
	defer xerr.Contextf(&err, "/: mkdir %q", name)

	var rev zodb.Tid
	ok := false

	if strings.HasPrefix(name, "@") {
		rev, err = zodb.ParseTid(name[1:])
		ok = (err == nil)
	}
	if !ok {
		return nil, eINVALf("not @rev")
	}

	// check to see if dir(rev) is already there
	root.revMu.Lock()
	_, already := root.revTab[rev]
	root.revMu.Unlock()

	if already {
		return nil, syscall.EEXIST
	}

	// not there - without revMu lock proceed to open @rev view of ZODB
	ctx := asctx(fctx)
//	zconnRev, err := root.zopenAt(ctx, rev)
	zconnRev, err := zopen(ctx, root.zdb, &zodb.ConnOptions{At: rev})
	if err != nil {
		return nil, err
	}

	// relock root and either mkdir or EEXIST if the directory was maybe
	// simultanously created while we were not holding revMu.
	root.revMu.Lock()
	_, already = root.revTab[rev]
	if already {
		root.revMu.Unlock()
//		zconnRev.Release()
		transaction.Current(zconnRev.txnCtx).Abort()
		return nil, syscall.EEXIST
	}

	// XXX -> newHead()
	revDir := &Head{
		Node:  newDefaultNode(),
		rev:   rev,
		zconn: zconnRev,
	}

	bfdir := &BigFileDir{
		Node:    newDefaultNode(),
		head:    revDir,
		fileTab: make(map[zodb.Oid]*BigFile),
	}
	revDir.bfdir = bfdir

	root.revTab[rev] = revDir
	root.revMu.Unlock()

	// mkdir takes filesystem treeLock - do it outside revMu.
	mkdir(root, name, revDir)
	mkdir(revDir, "bigfile", bfdir)
	// XXX + "at"

	return revDir.Inode(), nil
}


// bigopen opens BigFile corresponding to oid on head.zconn.
//
// A ZBigFile corresponding to oid is activated and statted.
//
// head.zconn must be locked.
func (head *Head) bigopen(ctx context.Context, oid zodb.Oid) (_ *BigFile, err error) {
	zconn := head.zconn
	defer xerr.Contextf(&err, "bigopen %s @%s", oid, zconn.At())

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(ctx, zconn.txnCtx)
	defer cancel()

	xzbf, err := zconn.Get(ctx, oid)
	if err != nil {
		switch errors.Cause(err).(type) {
		case *zodb.NoObjectError:
			return nil, eINVAL(err)
		case *zodb.NoDataError:
			return nil, eINVAL(err) // XXX what to do if it was existing and got deleted?
		default:
			return nil, err
		}
	}

	zbf, ok := xzbf.(*ZBigFile)
	if !ok {
		return nil, eINVALf("%s is not a ZBigFile", typeOf(xzbf))
	}

	// activate ZBigFile and keep it this way
	err = zbf.PActivate(ctx)
	if err != nil {
		return nil, err
	}
	defer func() {
		if err != nil {
			zbf.PDeactivate()
		}
	}()

	zbfSize, err := zbf.Size(ctx)
	if err != nil {
		return nil, err
	}

//	zconn.Incref()
	return &BigFile{
		Node:    newDefaultNode(),
		head:    head,
		zbf:     zbf,
		zbfSize: zbfSize,

		// XXX this is needed only for head/
		δFtail:  NewΔTailI64(zconn.At()),
		loading: make(map[int64]*blkLoadState),
	}, nil
}

// Close release all resources of BigFile.
func (f *BigFile) Close() error {
	f.zbf.PDeactivate()	// XXX f.head.zconn must locked
	f.zbf = nil

//	f.zconn.Release()
//	f.zconn = nil
	f.head = nil

	return nil
}

// /(head|<rev>)/bigfile/<bigfileX> -> Getattr serves stat.
func (f *BigFile) GetAttr(out *fuse.Attr, _ nodefs.File, _ *fuse.Context) fuse.Status {
	f.head.zconnMu.RLock()
	defer f.head.zconnMu.RUnlock()

	f.getattr(out)
	return fuse.OK
}

func (f *BigFile) getattr(out *fuse.Attr) {
	out.Mode = fuse.S_IFREG | 0444
	out.Size = uint64(f.zbfSize)
	// .Blocks
	// .Blksize

	// FIXME lastChange should cover all bigfile data, not only ZBigFile itself
	//mtime := f.δFtail[-1] || zbf.PSerial?
	lastChange := f.zbf.PSerial()	// XXX activate (if zbf becomes not activated during txn)

	mtime := lastChange.Time().Time
	out.SetTimes(/*atime=*/nil, /*mtime=*/&mtime, /*ctime=*/&mtime)
}


// /(head|<rev>)/bigfile/<bigfileX> -> Read serves reading bigfile data.
func (f *BigFile) Read(_ nodefs.File, dest []byte, off int64, fctx *fuse.Context) (fuse.ReadResult, fuse.Status) {
	f.head.zconnMu.RLock()
	defer f.head.zconnMu.RUnlock()

	zbf := f.zbf

	// cap read request to file size
	end := off + int64(len(dest))		// XXX overflow?
	if end > f.zbfSize {
		end = f.zbfSize
	}
	if end <= off {
		// XXX off >= size -> EINVAL? (but when size=0 kernel issues e.g. [0 +4K) read)
		return fuse.ReadResultData(nil), fuse.OK
	}

	// widen read request to be aligned with blksize granularity
	// (we can load only whole ZBlk* blocks)
	aoff := off - (off % zbf.blksize)
	aend := end
	if re := end % zbf.blksize; re != 0 {
		aend += zbf.blksize - re
	}
	dest = make([]byte, aend - aoff) // ~> [aoff:aend) in file

	// XXX better ctx = transaction.PutIntoContext(ctx, txn)
	ctx, cancel := xcontext.Merge(asctx(fctx), f.head.zconn.txnCtx)
	defer cancel()

	// read/load all block(s) in parallel
	wg, ctx := errgroup.WithContext(ctx)
	for blkoff := aoff; blkoff < aend; blkoff += zbf.blksize {
		blkoff := blkoff
		blk := blkoff / zbf.blksize
		wg.Go(func() error {
			δ := blkoff-aoff // blk position in dest
			//log.Infof("readBlk #%d dest[%d:+%d]", blk, δ, zbf.blksize)
			return f.readBlk(ctx, blk, dest[δ:δ+zbf.blksize])
		})
	}

	err := wg.Wait()
	if err != nil {
		log.Errorf("%s", err)	// XXX + /bigfile/XXX: read [a,b): -> ...
		return nil, fuse.EIO
	}

	return fuse.ReadResultData(dest[off-aoff:end-aoff]), fuse.OK
}

// readBlk serves Read to read 1 ZBlk #blk into destination buffer.
//
// see "7) when we receive a FUSE read(#blk) request ..." in overview.
//
// len(dest) == blksize.
func (f *BigFile) readBlk(ctx context.Context, blk int64, dest []byte) error {
	// XXX errctx?
	// XXX locking

	// check if someone else is already loading this block
	f.loadMu.Lock()
	loading, already := f.loading[blk]
	if !already {
		loading = &blkLoadState{
			ready:   make(chan struct{}),
		}
		f.loading[blk] = loading
	}
	f.loadMu.Unlock()

	// if it is already loading - just wait for it
	if already {
		select {
		case <-ctx.Done():
			return ctx.Err()

		case <-loading.ready:
			if loading.err == nil {
				copy(dest, loading.blkdata)
			}
			return loading.err
		}
	}

	// noone was loading - we became reponsible to load this block

	zbf := f.zbf
	blkdata, treepath, err := zbf.LoadBlk(ctx, blk)	// XXX -> +blkrevmax1
	loading.blkdata = blkdata
	loading.err = err
	close(loading.ready)

	_ = treepath
/*
	blkrevmax := zodb.Tid(0)
	for __, node := range treepath {
		node.PActivate()	// XXX err
		blkrevmax = max(_, node.PSerial())
		node.PDeactivate()

	}
*/

	// XXX before loading.ready?
	blkrevmax2, _ := f.δFtail.LastRevOf(blk, zbf.PJar().At())
	//revmax := min(blkrevmax1, blkrevmax2)
	revmax := blkrevmax2
	_ = revmax

/*
	// XXX remmapping - only if head.rev == 0
	// XXX -> own func?
	// XXX locking
	for _, mapping := range f.mappings {
		if revmax <= mapping.at || !mapping.blkrange.in(blk) {
			continue // do nothing
		}

		if mapping.pinned.Contains(blk) {
			continue // do nothing
		}

		rev = max(δFtail.by(blk) : _ <= mapping.at)

		// XXX vvv -> go
		client.remmap(mapping.addr[blk], file/@<rev>/data)
		mapping.pinned.Add(blk)


	}
*/


	// data loaded with error - cleanup .loading
	if loading.err != nil {
		f.loadMu.Lock()
		delete(f.loading, blk)
		f.loadMu.Unlock()
		return err
	}

	// data loaded ok
	copy(dest, blkdata)

	// store to kernel pagecache whole block that we've just loaded from database.
	// This way, even if the user currently requested to read only small portion from it,
	// it will prevent next e.g. consecutive user read request to again hit
	// the DB, and instead will be served by kernel from its pagecache.
	//
	// We cannot do this directly from reading goroutine - while reading
	// kernel FUSE is holding corresponging page in pagecache locked, and if
	// we would try to update that same page in pagecache it would result
	// in deadlock inside kernel.
	//
	// .loading cleanup is done once we are finished with putting the data into OS pagecache.
	// If we do it earlier - a simultaneous read covered by the same block could result
	// into missing both kernel pagecache (if not yet updated) and empty .loading[blk],
	// and thus would trigger DB access again.
	go func() {
		// XXX locking - invalidation must make sure this workers are finished.

		// XXX if direct-io: don't touch pagecache
		st := gfsconn.FileNotifyStoreCache(f.Inode(), blk*zbf.blksize, blkdata)

		f.loadMu.Lock()
		delete(f.loading, blk)
		f.loadMu.Unlock()

		if st == fuse.OK {
			return
		}

		// pagecache update failed, but it must not (we verified on startup that
		// pagecache control is supported by kernel). We can correctly live on
		// with the error, but data access will be likely very slow. Tell user
		// about the problem.
		log.Errorf("BUG: bigfile %s: blk %d: -> pagecache: %s  (ignoring, but reading from bigfile will be very slow)", zbf.POid(), blk, st)
	}()

	return nil
}





// FIXME groot/gfsconn is tmp workaround for lack of way to retrieve FileSystemConnector from nodefs.Inode
// TODO:
//	- Inode += .Mount() -> nodefs.Mount
//	- Mount:
//		.Root()		-> root Inode of the fs
//		.Connector()	-> FileSystemConnector through which fs is mounted
var groot   *Root
var gfsconn *nodefs.FileSystemConnector

// root of the filesystem is mounted here.
//
// we need to talk to kernel and lookup @<rev>/bigfile/<fid> before uploading
// data to kernel cache there. Referencing root of the filesystem via path is
// vulnerable to bugs wrt e.g. `mount --move` and/or mounting something else
// over wcfs. However keeping opened root fd will prevent wcfs to be unmounted,
// so we still have to reference the root via path.
var gmntpt string

// debugging
var gdebug = struct {
	// .wcfs/zhead opens
	// protected by groot.head.zconnMu
	zheadSockTab map[*FileSock]struct{}
}{}

func init() {
	gdebug.zheadSockTab = make(map[*FileSock]struct{})
}

// _wcfs_Zhead serves .wcfs/zhead opens.
type _wcfs_Zhead struct {
	nodefs.Node
}

func (zh *_wcfs_Zhead) Open(flags uint32, fctx *fuse.Context) (nodefs.File, fuse.Status) {
	// XXX check flags?
	sk := NewFileSock()
	sk.CloseRead()

	groot.head.zconnMu.Lock()
	defer groot.head.zconnMu.Unlock()

	gdebug.zheadSockTab[sk] = struct{}{}
	return sk.File(), fuse.OK
}

func main() {
	stdlog.SetPrefix("wcfs: ")
	//log.CopyStandardLogTo("WARNING") // XXX -> "DEBUG" if -d ?
	defer log.Flush()

	debug := flag.Bool("d", false, "debug")
	autoexit := flag.Bool("autoexit", false, "automatically stop service when there is no client activity")
	// XXX option to prevent starting if wcfs was already started ?

	flag.Parse()
	if len(flag.Args()) != 2 {
		log.Fatalf("Usage: %s [OPTIONS] zurl mntpt", os.Args[0])
	}
	zurl := flag.Args()[0]
	mntpt := flag.Args()[1]

	// debug -> precise t, no dates	(XXX -> always precise t?)
	if *debug {
		stdlog.SetFlags(stdlog.Lmicroseconds)
	}

	// open zodb storage/db/connection
	ctx := context.Background()	// XXX + timeout?
	zstor, err := zodb.OpenStorage(ctx, zurl, &zodb.OpenOptions{
		ReadOnly: true,
	})
	if err != nil {
		log.Fatal(err)
	}
	defer zstor.Close()

	zdb := zodb.NewDB(zstor)
	zhead, err := zopen(ctx, zdb, &zodb.ConnOptions{
		// we need zhead.cache to be maintained across several transactions.
		// see "3) for head/bigfile/* the following invariant is maintained ..."
		NoPool: true,
	})
	if err != nil {
		log.Fatal(err)
	}
	zhead.Cache().SetControl(&zodbCacheControl{})	// XXX +locking?

	// mount root + head/
	// XXX -> newHead()
	head := &Head{
		Node:  newDefaultNode(),
		rev:   0,
		zconn: zhead,
	}
	bfdir := &BigFileDir{
		Node:    newDefaultNode(),
		head:    head,
		fileTab: make(map[zodb.Oid]*BigFile),
	}
	head.bfdir = bfdir

	root := &Root{
		Node:   newDefaultNode(),
		zstor:  zstor,
		zdb:    zdb,
		head:   head,
		revTab: make(map[zodb.Tid]*Head),
	}

	opts := &fuse.MountOptions{
		FsName: zurl,
		Name:   "wcfs",

		DisableXAttrs: true, // we don't use
		Debug:         *debug,
	}

	fssrv, fsconn, err := mount(mntpt, root, opts)
	if err != nil {
		log.Fatal(err)
	}
	groot   = root		// FIXME temp workaround (see ^^^)
	gfsconn = fsconn	// FIXME ----//----
	gmntpt  = mntpt

	// we require proper pagecache control (added to Linux 2.6.36 in 2010)
	supports := fssrv.KernelSettings().SupportsNotify
	if !(supports(fuse.NOTIFY_STORE_CACHE) && supports(fuse.NOTIFY_RETRIEVE_CACHE)) {
		log.Fatalf("kernel FUSE does not support pagecache control")
	}

	// add entries to /
	mkdir(root, "head", head)
	mkdir(head, "bigfile", bfdir)
	mkfile(head, "at", NewSmallFile(head.readAt))   // TODO mtime(at) = tidtime(at)
	// XXX ^^^ invalidate cache or direct io XXX no FOPEN_KEEP_CACHE -> direct io

	// for debugging/testing
	_wcfs := newDefaultNode()
	mkdir(root, ".wcfs", _wcfs)
	mkfile(_wcfs, "zurl", NewStaticFile([]byte(zurl)))

	// .wcfs/zhead - special file channel that sends zhead.at.
	//
	// If a user opens it, it will start to get tids of through which
	// zhead.at was, starting from the time when .wcfs/zhead was opened.
	// There can be multiple openers. Once opened, the file must be read,
	// as wcfs blocks waiting for data to be read before XXX.
	mkfile(_wcfs, "zhead", &_wcfs_Zhead{
		Node: newDefaultNode(),
	})

	// XXX place = ok?
	// XXX ctx = ok?
	// XXX wait for zwatcher shutdown.
	go root.zwatcher(ctx)

	// TODO handle autoexit
	_ = autoexit

	// serve client requests
	fssrv.Serve()	// XXX Serve returns no error
}
